/*******************************************************************************
* Data: 04 kwietnia 2005
* LN_GET_VALUE - dodanie do aktualizacji stosu warunku: STAN = 0
*******************************************************************************/

/*******************************************************************************
* Baza danych do obsugi licznikw.
* Podczenie z poziomu Buildera C++:
*    TransIsolation = tiRepeatableRead  (warunek konieczny poprawnego dziaania)
*    Params :
*      WAIT ON LOCKS=FALSE              (moe by take TRUE)
*      LANGDRIVER=Paradox Polish 852   
* Procedury pamitane musz by wywoywane w obrbie podanej jawnie transakcji.
* W przypadku wystpienia bdu zmiany dokonane przez procedur musz zosta
* wycofane (Rollback). Inne postpowanie moe by przyczyn powtrzenia numeracji.
* Przykad:
*    try {
*       dbLICZNIKI->StartTransaction();
*       spGetValue->Prepare();
*       spGetValue->ExecProc();
*       dbLICZNIKI->Commit();
*       ... = spGetValue->ParamByName('WARTOSC')->AsInteger; // zwrcona warto
*    }
*    catch(Exception& exc) {
*        dbLICZNIKI->Rollback();
*    }
* Modyfikowanie tabel tylko za pomoc procedur wbudowanych.
* Wartoci licznika nale do zbioru liczb cakowitych nieujemnych.
* -----------------------------------------------------------------------------
* LN_CHECK_CODE (IdSystemu INTEGER, Rok INTEGER, KodLicznika CHARACTER(6))
*    - sprawdza, czy istnieje licznik o podanym kodzie (0-nie, 1-tak)
* LN_GET_VALUE (IdSystemu INTEGER, Rok INTEGER, KodLicznika CHARACTER(6))
*    - pobierz warto z licznika o podanym kodzie (zgasza wyjtek w przypadku bdu),
* LN_APPLY_VALUE (IdSystemu INTEGER, Rok INTEGER, KodLicznika CHARACTER(6), Wartosc INTEGER)
*    - zatwierd podan warto, pobran z licznika o podanym kodzie (zgasza wyjtek w 
*      przypadku bdu),
* LN_CANCEL_VALUE (IdSystemu INTEGER, Rok INTEGER, KodLicznika CHARACTER(6), Wartosc INTEGER)
*    - wycofaj podan warto do ponownego wykorzystania, pobran z licznika o
*      podanym kodzie (zgasza wyjtek w przypadku bdu),
* LN_CLEAR_STACK (IdSystemu INTEGER, Rok INTEGER, KodLicznika CHARACTER(6))
*    - usu ze stosu wskazanego licznika wszystkie zapisy (zgasza wyjtek w przypadku bdu),
* LN_SET_TO (IdSystemu INTEGER, Rok INTEGER, KodLicznika CHARACTER(6), Wartosc INTEGER)
*    - ustawia warto wskazanego licznika i czyci jego stos (zgasza wyjtek w przypadku
*      bdu).
*******************************************************************************/

SET NAMES WIN1250;

/******
*   LN_LICZNIKI (0;N) ---> (1;1) LN_STOS
******/

/******
* LN_LICZNIKI
*   tabela zawierajca definicje licznikw i ich gwne wartoci; licznik jest
*   identyfikowany przez kod systemu i znakowy kod licznika w systemie
******/
CREATE TABLE LN_LICZNIKI
(   
   ID_SYSTEMU   INTEGER NOT NULL,
   ROK          INTEGER NOT NULL,
   KOD          VARCHAR(2) NOT NULL,
   NAZWA        VARCHAR(50) COLLATE WIN1250,
   GLOWNY       INTEGER,
      CONSTRAINT LN_PK_LICZNIKI PRIMARY KEY (ID_SYSTEMU, ROK, KOD)
);

/******
* LN_STOS
*   tabela zawiera stosy licznikw zdefiniowanych w tabeli LN_LICZNIKI;
*   na stosie odkadane s wartoci pobrane z licznikw (STAN = 1); zatwierdzenie
*   danej wartoci powoduje usunicie jej ze stosu (DELETE); wycofana warto jest
*   zaznaczana do ponownego wykorzystania (STAN = 0).
******/
CREATE TABLE LN_STOS
(  
   ID_SYSTEMU   INTEGER NOT NULL,
   ROK          INTEGER NOT NULL,
   KOD          VARCHAR(2) NOT NULL, 
   WARTOSC      INTEGER NOT NULL,
   STAN         INTEGER,
      CONSTRAINT LN_PK_STOS PRIMARY KEY (ID_SYSTEMU, ROK, KOD, WARTOSC)
);

/*********************************************************************************************
* INTEGRALNO REFERENCYJNA
*********************************************************************************************/

ALTER TABLE LN_STOS
   ADD CONSTRAINT LN_REL_GLW2STS FOREIGN KEY (ID_SYSTEMU, ROK, KOD)
      REFERENCES LN_LICZNIKI (ID_SYSTEMU, ROK, KOD)
   ON UPDATE CASCADE
   ON DELETE CASCADE;

/****
* Wizy
****/

ALTER TABLE LN_LICZNIKI
   ADD CONSTRAINT LN_CST_GLW_GLOWNY CHECK( GLOWNY IS NOT NULL AND GLOWNY >= 0 );
   
ALTER TABLE LN_STOS
   ADD CONSTRAINT LN_CST_STS_WARTOSC CHECK( WARTOSC >= 0 ),
   ADD CONSTRAINT LN_CST_STS_STAN CHECK( STAN IS NOT NULL AND (STAN = 0 OR STAN = 1) );


CREATE EXCEPTION LN_EXC_INVALID_ID    'Licznik o podanym kodzie nie istnieje';
CREATE EXCEPTION LN_EXC_INVALID_VALUE 'Prba zatwierdzenia lub wycofania nieznanej wartoci';
CREATE EXCEPTION LN_EXC_NO_VALUE      'Warto licznika nie zostaa pobrana';


SET TERM !! ;

/******
* LN_CHECK_CODE
* Argumenty:
*    IdSystemu - identyfikator systemu,
*    KodLicznika - kod licznika
* sprawdza, czy istnieje licznik o podanym kodzie (w ramach danego systemu)
******/
CREATE PROCEDURE LN_CHECK_CODE (IdSystemu INTEGER, Rok INTEGER, KodLicznika VARCHAR(2))
RETURNS (KodBledu INTEGER) AS
   DECLARE VARIABLE GlWartosc INTEGER;
   BEGIN

      KodBledu = 0;

      SELECT glowny FROM LN_LICZNIKI 
      WHERE id_systemu = :IdSystemu AND rok = :Rok AND kod = :KodLicznika
      INTO :GlWartosc;

      /*
      * jeeli zmienna GlWartosc jest rwna NULL, to przekazany zosta
      * kod nie istniejcego licznika
      */
      IF (GlWartosc IS NOT NULL) THEN
         KodBledu = 1;

   END !!

/******
* LN_GET_VALUE
* Argumenty:
*    IdSystemu - identyfikator systemu,
*    KodLicznika - kod licznika
* Zwraca kolejn warto z licznika o podanym kodzie
******/
CREATE PROCEDURE LN_GET_VALUE (IdSystemu INTEGER, Rok INTEGER, KodLicznika VARCHAR(2))
RETURNS (Wartosc INTEGER) AS
   DECLARE VARIABLE GlWartosc INTEGER;
   DECLARE VARIABLE Stos INTEGER;
   BEGIN

      Wartosc = -1;

      SELECT glowny FROM LN_LICZNIKI 
      WHERE id_systemu = :IdSystemu AND rok = :Rok AND kod = :KodLicznika
      INTO :GlWartosc;

      /*
      * jeeli zmienna GlWartosc jest rwna NULL, to przekazany zosta
      * kod nie istniejcego licznika
      */
      IF (GlWartosc IS NULL) THEN
         EXCEPTION LN_EXC_INVALID_ID;

      ELSE BEGIN

         /*
         * podaj najmniejsz warto przechowywan na stosie
         */
         SELECT MIN(wartosc) FROM LN_STOS
         WHERE id_systemu = :IdSystemu AND rok = :Rok AND kod = :KodLicznika AND stan = 0
         INTO :Stos;

         IF ((Stos IS NOT NULL) AND (Stos > 0)) THEN BEGIN
            UPDATE LN_STOS SET stan = 1
            WHERE id_systemu = :IdSystemu AND rok = :Rok AND kod = :KodLicznika AND stan = 0 AND wartosc = :Stos;

            Wartosc = Stos;

         END
         ELSE BEGIN
            /*
            * zwikszam gwn warto licznika (na stosie nie ma wolnych
            * numerw wybranego licznika)
            */
            GlWartosc = GlWartosc + 1;

            UPDATE LN_LICZNIKI SET glowny = :GlWartosc
            WHERE id_systemu = :IdSystemu AND rok = :Rok AND kod = :KodLicznika;

            INSERT INTO LN_STOS(id_systemu,rok,kod,wartosc,stan) 
            VALUES (:IdSystemu, :Rok, :KodLicznika, :GlWartosc, 1);

            Wartosc = GlWartosc;

         END

         /****
         * przechwycenie bdw:
         * -615 : rekord zablokowany (dotyczy: UPDATE)
         * -803 : powtrzony klucz gwny (dotyczy: INSERT)
         * -913 : deadlock
         ****/
         WHEN SQLCODE -913, SQLCODE -803, SQLCODE -615 DO 
         BEGIN
            Wartosc = -1;
            EXCEPTION LN_EXC_NO_VALUE;
         END

      END

   END !!

/*
* LN_CANCEL_VALUE
* Argumenty:
*    IdSystemu - identyfikator systemu,
*    KodLicznika - kod licznika,
*    Wartosc - warto zwrcona przez procedur LK_GET_VALUE
* Wycofuje warto pobran z licznika. Moe zosta ponownie uyta.
*/
CREATE PROCEDURE LN_CANCEL_VALUE (IdSystemu INTEGER, Rok INTEGER, KodLicznika VARCHAR(2), Wartosc INTEGER)
RETURNS (KodBledu INTEGER) AS
   BEGIN

      /****
      * sprawdzam, czy podany zosta poprawny kod licznika
      ****/
      EXECUTE PROCEDURE LN_CHECK_CODE :IdSystemu, :Rok, :KodLicznika RETURNING_VALUES :KodBledu;

      IF (KodBledu = 0) THEN
         EXCEPTION LN_EXC_INVALID_ID;

      /****
      * nastpny SELECT nie nadpisze wartoci zmiennej KodBledu, jeeli
      * nie zostanie znaleziony rekord speniajce podane warunki; std
      * konieczno przypisania zmiennej KodBledu wartoci NULL, aby
      * porwnanie z NULL miao sens
      ****/
      KodBledu = NULL;

      /****
      * jeeli kod bdu jest rwny NULL, to nie zostaa na stosie
      * znaleziona przekazana warto zwizana ze wskazanym licznikiem
      ****/
      SELECT wartosc FROM LN_STOS
      WHERE id_systemu = :IdSystemu AND rok = :Rok AND kod = :KodLicznika AND
            wartosc = :Wartosc AND stan = 1
      INTO :KodBledu;

      IF (KodBledu IS NULL) THEN BEGIN
         KodBledu = 0;
         EXCEPTION LN_EXC_INVALID_VALUE;
      END

      KodBledu = 0;

      UPDATE LN_STOS SET stan = 0
      WHERE id_systemu = :IdSystemu AND rok = :Rok AND kod = :KodLicznika AND
            wartosc = :Wartosc;

      KodBledu = 1;

   END !!

/*
* LN_APPLY_VALUE
* Argumenty:
*    IdSystemu - identyfikator systemu,
*    KodLicznika - kod licznika,
*    Wartosc - warto zwrcona przez procedur LK_GET_VALUE
* Zatwierdza warto pobran z licznika.
*/
CREATE PROCEDURE LN_APPLY_VALUE (IdSystemu INTEGER, Rok INTEGER, KodLicznika VARCHAR(2), Wartosc INTEGER)
RETURNS (KodBledu INTEGER) AS
   BEGIN

      /*
      * sprawdzam, czy podany zosta poprawny kod licznika
      */
      EXECUTE PROCEDURE LN_CHECK_CODE :IdSystemu, :Rok, :KodLicznika RETURNING_VALUES :KodBledu;

      IF (KodBledu = 0) THEN
         EXCEPTION LN_EXC_INVALID_ID;

      /****
      * nastpny SELECT nie nadpisze wartoci zmiennej KodBledu, jeeli
      * nie zostanie znaleziony rekord speniajce podane warunki; std
      * konieczno przypisania zmiennej KodBledu wartoci NULL, aby
      * porwnanie z NULL miao sens
      ****/
      KodBledu = NULL;

      /****
      * jeeli kod bdu jest rwny NULL, to nie zostaa na stosie
      * znaleziona przekazana warto zwizana ze wskazanym licznikiem
      ****/
      SELECT wartosc FROM LN_STOS
      WHERE id_systemu = :IdSystemu AND rok = :Rok AND kod = :KodLicznika AND wartosc = :Wartosc AND stan = 1
      INTO :KodBledu;

      IF (KodBledu IS NULL) THEN BEGIN
         KodBledu = 0;
         EXCEPTION LN_EXC_INVALID_VALUE;
      END

      KodBledu = 0;

      DELETE FROM LN_STOS 
      WHERE id_systemu = :IdSystemu AND rok = :Rok AND kod = :KodLicznika AND wartosc = :Wartosc;

      KodBledu = 1;

   END !!

/*
* LN_SET_TO
* Argumenty:
*    IdSystemu - identyfikator systemu,
*    KodLicznika - kod licznika,
*    Wartosc - warto, na jak ma by ustawiony licznik; jeeli argument
*              jest mniejszy od zera, to licznik jest zerowany.
* Ustawia licznik na podan warto. Usuwa zapisy ze stosu.
*/
CREATE PROCEDURE LN_SET_TO (IdSystemu INTEGER, Rok INTEGER, KodLicznika VARCHAR(2), Wartosc INTEGER)
RETURNS (KodBledu INTEGER) AS
   BEGIN

      /*
      * warto licznika nie moe by mniejsza od zera
      */
      IF (Wartosc < 0) THEN
         Wartosc = 0;

      /*
      * sprawdzam, czy podany zosta poprawny kod licznika
      */
      EXECUTE PROCEDURE LN_CHECK_CODE :IdSystemu, :Rok, :KodLicznika RETURNING_VALUES :KodBledu;

      IF (KodBledu = 0) THEN
         EXCEPTION LN_EXC_INVALID_ID;

      KodBledu = 0;

      /*
      * pierwszy krok - usuwane s wszystkie wartoci ze stosu
      */
      DELETE FROM LN_STOS WHERE id_systemu = :IdSystemu AND rok = :Rok AND kod = :KodLicznika;

      /*
      * drugi krok - ustawiana jest warto gwna licznika
      */
      UPDATE LN_LICZNIKI SET glowny = :Wartosc 
      WHERE id_systemu = :IdSystemu AND rok = :Rok AND kod = :KodLicznika;

      KodBledu = 1;

   END !!

/*
* LN_CLEAR_STACK
* Argumenty:
*    IdSystemu - identyfikator systemu,
*    KodLicznika - kod licznika,
* Czyci stos licznika.
*/
CREATE PROCEDURE LN_CLEAR_STACK (IdSystemu INTEGER, Rok INTEGER, KodLicznika VARCHAR(2))
RETURNS (KodBledu INTEGER) AS
   BEGIN

      /*
      * sprawdzam, czy podany zosta poprawny kod licznika
      */
      EXECUTE PROCEDURE LN_CHECK_CODE :IdSystemu, :Rok, :KodLicznika RETURNING_VALUES :KodBledu;

      IF (KodBledu = 0) THEN
         EXCEPTION LN_EXC_INVALID_ID;


      KodBledu = 0;

      /*
      * usuwane s zapisy ze stosu
      */
      DELETE FROM LN_STOS WHERE id_systemu = :IdSystemu AND rok = :Rok AND kod = :KodLicznika;

      KodBledu = 1;

   END !!


SET TERM ; !!

